'******************************************************************
'*  Name    : SnapShot.inc                                        *
'*  Author  : George Towler                                       *
'*  Date    : see History                                         *
'*  Version : see VersionHigh/VersionLow                          *
'*  Compiler: 3.6.2.8                                             *
'*  Notes   : Functionally this behaves as a compile time         *
'*            breakpoint with variable display and optional wait. *
'*                                                                *
'*      PIC device is client.  PC is host.                        *
'*                                                                *
'* Client's data memory is formed into packets and send serially. *
'* Packets are fixed length and are of several types.             *
'* All packets start with STX and end with CRC.                   *
'* The PC host is configurable (via SnapShot.ini) To any standard *
'*   baud rate and port number.                                   *
'* Major version number should be raised on both whenever         *
'*   changes to either client or host would break the system.     *
'*                                                                *
'* Note! SerIn is used even though RsIn generates less code       *
'*   because RsIn uses a fixed timeout                            *
'*                                                                *
'******************************************************************
'History
' 1.0  06/01/19  removed any reference To D_DumpingLED
' 0.9  29/12/18  changed way D_DumpLED declared, minor doc changes, corrected typo, D_UseHardSerial being phased out
'                reordered subroutines
'                rewrote D_ScanForNull - only test for $FF once
'                Dev Id and Rev Id on new devices. change to EndData format
'                Workaround for 18F26K42 & 18F27K42 at 9600 baud
'                HexTag position changed, HexTag found by label not position
'                Heap support removed, it's bank switching is fucked in 3.6.1.6
'                Assembler macros replaced with preprocessor equivalents (for 18F26K42 and  18F27K42)
'                added D_LimitEEPROM to bypass problems with word addresses on EEPROM for 18F26K42 and 18F27K42
' 0.8  26/05/18  simplified desired baud rate definition
'                check is valid
'                extra baud rates added for bit bang
'                added D_SendPort option
'                changed D_RxTimeOut and remmed D_PacketDelay from code
'                added D_HardSerialNum option for PIC with multiple UARTS
'                added config for 18F25K50, 18F26K22 and 18F26K42 (only partial functional)
'                use _core to set default D_CheckNulls and D_UseCRC
'                detect CRead fail if HexTag CData not present for 16F
' 0.7  17/12/17  IDLOCS replaced with ROM based SourceID
' 0.6  01/11/17  CRC code and header addded to main body to simplify distribution
'                IDLOCS added
'                DevID automatically detected
'                changed CRC8 variable names to reduce risk of reuse by PUT
'                Undersize NULL/FF optimisation enabled
'                TBLPTRU=0 after use and control EECON1 to read device ID - unbork string handling
'                Heap support added
'                added data for HardSerial @ 9600
'                D_ResetFlags used symbolically
'                Source verification added
' 0.4  07/07/17  converted EndIf to End If
'                made System Variable Save/Restore more bulletproof
'                made address of DevID Device dependant
'                added D_TempWord to prevent involvment PBP_VAR sys var
' 0.3  15/05/17  resolved D_AddressEnd corruption
'                changed the way dump LED defined
'                Added packet len to all inbound packet types to ease pharsing
'                Send EEPROM content
'                Uses WaitNak to filter timeout in WaitSync phase
'                Change comms timeout from 1 sec to 0.25
'                16F support added but with edit disabled
'                D_UseCRC added to skip CRC calc - not recommended
'                Reset PIC option added (18F only)
'                Modified EndData packet to include return address and DeviceID (18F only)
'                D_CheckNulls added to remove contiguous nulls optimisation
'                Automated port data
' 0.2  29/01/17  Modified EndData packet to include start and end address
' 0.1  27/01/17  First release

  Symbol D_VersionHigh  = 1           ' major ver to PC - changes which break the system
  Symbol D_VersionLow   = 0           ' minor ver to PC - minor changes which do not

  $ifndef False
    $define False 0
  $endif
  $ifndef True
    $define True 1
  $endif

  $if _flash = 0 Or _code < 4096
    $error _device " is not a suitable candidate for Snapshot"
  $endif

  $if D_SendEEPROM = True
    $if _device = _18F27K42
      $if D_LimitEEPROM = False And D_SendEEPROM = True
        $error "For this device D_LimitEEPROM must be True (untill word address read/write EEPROM is resolved)"
      $endif
    $endif
  $endif

  '**************************** Default Client Settings ****************** (add settings to client application to override)


  $ifndef D_SerialNum           ' unless specified set UART appropriatly
    $if _uart = 1
      $define D_SerialNum 1     ' Use first UART if only one
    $elseif _uart > 1
      $define D_SerialNum 2     ' Use second UART if 2 or more
    $elseif _uart > 0
      $define D_SerialNum 0     ' Use second UART if 2 or more
    $endif
  $endif

  $if D_SerialNum > _uart
    $error "Incorrect setting for D_SerialNum - nonexistant USART"  ' check for physical USART
  $endif

  $ifndef D_BaudRate
    $if D_SerialNum = 0
      $define D_BaudRate 9600         ' default for bit bang
    $else
      $define D_BaudRate 115200       ' default for real UART
    $endif
  $endif

  $ifndef D_SendPort
    $define D_SendPort True           ' default to True
  $endif

  $ifndef D_UseCRC
    $if _core = 16
      $define D_UseCRC True           ' 18F default to compute CRC
    $else
      $define D_UseCRC False          ' 16F default to use placeholder instead of CRC
    $endif
  $endif

  $ifndef D_CheckNulls
    $if _core = 16
      $define D_CheckNulls True       ' 18F default to CheckNulls
    $else
      $define D_CheckNulls False      ' 16F default to bypass CheckNulls
    $endif
  $endif

  '**************************** end of Default Client Settings ******************

  $if _eeprom
    $ifndef D_SendEEPROM
      $define D_SendEEPROM True       ' if available default to sending EEPROM contents
    $endif
  $else
    $ifdef D_SendEEPROM               '  +
      $undef D_SendEEPROM             '  |    **** don't change - check for PIC without EEPROM ****
    $endif                            '  |
    $define D_SendEEPROM False        '  +
  $endif

  Symbol D_PacketLen    = 32                          ' <------ **** set fixed packet size *****
  Symbol D_PortPad      = D_PacketLen - 13            ' STX, Type, 5xTRIS, 5xLAT, CRC
  Symbol D_NullPad      = D_PacketLen - 6             ' STX, Type, AddressL, AddressH, Len, CRC
  Symbol D_FFPad        = D_PacketLen - 6             ' STX, Type, AddressL, AddressH, Len, CRC
  Symbol D_EndDataPad14 = D_PacketLen - 25            ' STX, Type, DumpID, VerH, VerL, WaitFlag, StartAdd, EndAdd, SourceID x14, pad x7,CRC                     25
  Symbol D_EndDataPad16 = D_PacketLen - 30            ' STX, Type, DumpID, VerH, VerL, WaitFlag, StartAdd, EndAdd, SourceID x14, Devid x2, Flags, pad x4, CRC    28
                                                      '  if necessary can combine WaitFlag and Flags, saving 1 byte
  Symbol D_DataOverH    = 6                           ' STX, Type, Len, AddressL, AddressH, CRC
  Symbol D_EEPROMOverH  = 6                           ' STX, Type, Len, AddressL, AddressH, CRC
  Symbol D_SFROverH     = 5                           ' STX, Type, AddressL, AddressH, CRC
  Symbol D_DataBlock    = D_PacketLen - D_DataOverH   ' size of data payload in data packet
  Symbol D_EEPROMBlock  = D_PacketLen - D_EEPROMOverH ' size of EEPROM payload in data packet
  Symbol D_SFRBlock     = D_PacketLen - D_SFROverH    ' size of EEPROM payload in data packet

  Symbol D_TimeOut      = 250                         ' Timeout (in msec) waiting For ACK/NAK from PC
  Symbol D_PacketDelay  = 100                         ' delay (in usecs) after STX                          remmed from code 23/03/18
  Symbol D_RxTimeOut    = 20                          ' Timeout (in msec) waiting For packet data from PC   was 10 23/03/18
  Symbol D_Retry        = 10                          ' if count exceeded, revert to SYNC condition
  Symbol D_Pad          = 0                           ' pad character
  Symbol D_ResetFlagH   = $CC
  Symbol D_ResetFlagL   = $55

  ' Message types
  Symbol D_MT_Data       = 7                          ' data block 1 thro D_DataBlock bytes                       PIC > PC
  Symbol D_MT_NullBlock  = 8                          ' block of contiguous NULLs in RAM                          PIC > PC
  Symbol D_MT_EndData    = 9                          ' all data sent                                             PIC > PC
  Symbol D_MT_ROMEndData = 10                         ' last 2 or 4 bytes in used ROM and address                 PC > PIC
  Symbol D_MT_Port       = 11                         ' port data TRIS and LAT                                    PIC > PC
  Symbol D_MT_Modify     = 12                         ' modify PIC RAM                                            PC > PIC
  Symbol D_MT_Reset      = 13                         ' reset the PIC                                             PC > PIC
  Symbol D_MT_EEPROM     = 14                         ' data block form EEPROM                                    PIC > PC
  Symbol D_MT_FFBlock    = 15                         ' block of contingous FF in EEPROM (EEPROM clears to $FF)   PIC > PC
  Symbol D_MT_SFR        = 16                         ' data block from SFR

  ' message preamble
  Symbol D_STX          = 2                           ' start of packet  %10010010   146  $92

  ' Acknowledgements from PC
  Symbol D_WaitNak      = "a"         ' $61  97       ' negative response from PC on Wait Sync             '10000001
  Symbol D_WaitAck      = "b"         ' $62  98       ' positive response from PC On Wait Sync             '01000010
  Symbol D_SyncNak      = "g"         ' $67 103       ' negative response from PC on Sync                  '00100100
  Symbol D_SyncAck      = "h"         ' $68 104       ' positive response from PC on Sync                  '00011000
  Symbol D_Nak          = "e"         ' $65 101       ' negative response from PC on packet                '01111110
  Symbol D_Ack          = "f"         ' $66 102       ' positive response from PC on packet                '10111101

  ' Acknowledgements from PIC
  Symbol D_NotSupported = "i"         ' $69 105       ' response from PIC to unsupported InBoundPacket. PIC also sends D_Ack and D_Nak

  ' Syncronization from PIC
  Symbol D_Sync         = "c"         ' $63  99       ' initially send sync every second till Ack'ed       '11011011
  Symbol D_WaitSync     = "d"         ' $64 100       ' while waiting to restart send waitsync till Ack'ed '11100111


  ' Proton system variables 29                         c-common s-bitbang h-hrsin o-hrsout
  Dim BPF              As Byte System       'c
  Dim BPFH             As Byte System       's   o
  Dim GPR              As Byte System       's
  Dim GEN              As Byte System       'c   h
  Dim GENH             As Byte System       'c   h
  Dim GEN2             As Byte System
  Dim GEN2H            As Byte System
  Dim GEN3             As Byte System       's
  Dim GEN3H            As Byte System       's
  Dim GEN4             As Byte System
  Dim GEN4H            As Byte System
  Dim PP0              As Byte System       'c   h
  Dim PP0H             As Byte System       'c   h
  Dim PP1              As Byte System       'c   h
  Dim PP1H             As Byte System       'c   h
  Dim PP2              As Byte System
  Dim PP2H             As Byte System
  Dim PP3              As Byte System       's
  Dim PP3H             As Byte System       's
  Dim PP4              As Byte System       's
  Dim PP4H             As Byte System
  Dim PP5              As Byte System       's
  Dim PP5H             As Byte System       's
  Dim PP6              As Byte System
  Dim PP6H             As Byte System
  Dim PP7              As Byte System       'c
  Dim PP7H             As Byte System       'c
  Dim PP8              As Byte System
  Dim PP8H             As Byte System
  Dim Pbp#Var0         As Byte System

  Dim D_AddressBack    As Word                        ' Do not move this Variable, memory backup of start address in case data must be resent
  Dim D_AddressStart   As Word                        ' set in code under test
  Dim D_AddressEnd     As Word                        ' a constant but not settable with Proton
  Dim D_Address        As Word System                 '
  Dim D_TempWord       As Word                        ' used to prevent involvment PBP_VAR sys var

  Dim D_DataLen        As Byte                        ' length of data payload in packet
  Dim D_Data           As Byte System                 ' working variable
  Dim D_Counter        As Byte                        ' working variable
  Dim D_RetryCount     As Byte                        ' if count exceeded revert to sync condition
  Dim D_DumpID         As Byte                        ' used to syncronise with source display on PC
  Dim D_WaitFlag       As Bit                         ' 1=Wait for host  0=Return at once
  Dim D_RxFailFlag     As Bit                         ' 1=rx packet fail  0=appearantly good rx packet
  Dim D_RomGoodFlag    As Bit                         ' source verification data looks good

  $if _core = 16
    Dim D_MessageType  As Byte                        ' these variables are only used by D_InBoundPacket (18F only)
    Dim D_Data1        As Byte                        ' modify data
    Dim D_Data2        As Byte                        ' modify data
    Dim D_Data3        As Byte                        ' modify data
    Dim D_Data4        As Byte                        ' modify data
  $endif

  Dim D_WregBack       As Byte                        ' System vars used for hard serial
  Dim D_aBPF           As Byte
  Dim D_aGEN           As Byte
  Dim D_aGENH          As Byte
  Dim D_aPP0           As Byte
  Dim D_aPP0H          As Byte
  Dim D_aPP1           As Byte
  Dim D_aPP1H          As Byte
  Dim D_aPP7           As Byte
  Dim D_aPP7H          As Byte
  Dim D_Pbp#Var0       As Byte

  'jic
'  Dim D_aPP2           As Byte
'  Dim D_aPP2H          As Byte
'  Dim D_aPP4H          As Byte
'  Dim D_aPP6           As Byte
'  Dim D_aPP6H          As Byte
'  Dim D_aPP8           As Byte
'  Dim D_aPP8H          As Byte

  $if D_SerialNum = 0
    Dim D_aBPFH        As Byte                        ' Additional Sys vars used for serial bitbang
    Dim D_aGPR         As Byte
    Dim D_aGEN3        As Byte
    Dim D_aGEN3H       As Byte
    Dim D_aPP3         As Byte
    Dim D_aPP3H        As Byte
    Dim D_aPP4         As Byte
    Dim D_aPP5         As Byte
    Dim D_aPP5H        As Byte
  $endif

  $if D_UseCRC = True
    Symbol D_CRC8POLY = $18

    Dim D_CRC8_Comp                As Byte
    Dim D_CRC8_bit_counter         As Byte
    Dim D_CRC8_feedback_bit        As Bit
  $endif


'*************************** Device specific section - change to suit ************************************

  $if _device = _16F690 Or _device = _18F14K50
    Symbol PGD = PORTA.0
    Symbol PGC = PORTA.1
  $elseif _device = _16F886 Or _device = _18F2550 Or _device = _18F4550 Or _device = _18F25K50
    Symbol PGD = PORTB.7
    Symbol PGC = PORTB.6
  $else
    Symbol PGD = PORTB.7                           ' best quess for unspecified device
    Symbol PGC = PORTB.6
  $endif


  ' Any additional setup for particular device and/or baud rate

  $if D_SerialNum <> 0
    $if _device = _16F690 Or _device = _16F886
      ' for reason unclear it is necessary to ad Baud Gen values
      Declare Hserial_TXSTA = 36                   ' same for all
      Declare Hserial_RCSTA = 144                  ' same for all
      Declare Hserial_Clear = On                   ' same for all
      BAUDCTL.3             = 1                    ' Enable the 16 bit Baud Rate Generator
      $if D_BaudRate = 115200
        Declare Hserial_Baud = 115200
        SPBRG                = 16                  ' Baud Rate Generator Low Byte Value
        SPBRGH               = 0                   ' Baud Rate Generator High Byte Value
      $elseif D_BaudRate = 57600
        Declare Hserial_Baud = 57600
        SPBRG                = 34                  ' Baud Rate Generator Low Byte Value
        SPBRGH               = 0                   ' Baud Rate Generator High Byte Value
      $else
        Declare Hserial_Baud = 9600
        SPBRG                = 207                 ' Baud Rate Generator Low Byte Value
        SPBRGH               = 0                   ' Baud Rate Generator High Byte Value
      $endif

    $elseif _device = _18F2550 Or _device = _18F14K50 Or _device = _18F4550 Or _device = _18F25K50
      $if D_BaudRate = 115200
        Declare Hserial_Baud = 115200
        Declare Hserial_TXSTA = 36
      $elseif D_BaudRate = 57600
        Declare Hserial_Baud = 57600
        Declare Hserial_TXSTA = 36
      $else
        Declare Hserial_Baud = 9600                      ' default to 9600
        Declare Hserial_TXSTA = 32
      $endif
      Declare Hserial_RCSTA = 144                        ' same for all
      Declare Hserial_Clear = On

    $elseif _device = _18F26K22 Or _device = _18F27K42 Or _device = _18F27K40
      $if D_SerialNum = 1
        Declare Hserout1_Pin PORTC.6                     ' NOTE! devices with two UARTS eg. 18F26K22 etc
        Declare Hserin1_Pin PORTC.7
        Declare Hserial1_RCSTA = 144
        Declare Hserial1_Clear = On
      $else
        $if _device = _18F26K22
          Declare Hserout2_Pin PORTB.6
          Declare Hserin2_Pin PORTB.7
        $else
          Declare Hserout2_Pin PORTC.1
          Declare Hserin2_Pin PORTC.0
        $endif
        Declare Hserial2_RCSTA = 144                     ' same for all 2nd. UART
        Declare Hserial2_Clear = On
      $endif

      $if D_BaudRate = 115200
        $if D_SerialNum = 1
          Declare Hserial1_Baud  = 115200                ' used if D_BaudRate is undefined
          Declare Hserial1_TXSTA = 36
        $else
          Declare Hserial2_Baud  = 115200
          Declare Hserial2_TXSTA = 36
        $endif
      $elseif D_BaudRate = 57600
        $if D_SerialNum = 1
          Declare Hserial1_Baud  = 57600
          Declare Hserial1_TXSTA = 36
        $else
          Declare Hserial2_Baud  = 57600
          Declare Hserial2_TXSTA = 36
        $endif
      $else
        $if D_SerialNum = 1
          $if _device = _18F26K22 Or _device = _18F27K40
            Declare Hserial1_Baud  = 9600                ' if neither of above, which is unfortunate for 18F27K42
            Declare Hserial1_TXSTA = 32
          $else
            ' ***** work around for K42 - remove when issue resolved
            U1CON0 = %00110000                           ' BRGS=0 ABDEN=0 TXEN=1 RXEN=1 Mode=0000 (8bit async)
            U1CON1 = %10000000                           ' SerialEn=1 unimp=00 WUE=0 RXBIMD=0 unimp=0 BRKOVR=0 SENDB=0
            While U1FIFO.3 = 0                           ' RXIDL = bit 3 1 when idle
            Wend                                         '
            U1BRGH = 1                                   ' if idle set for 9600 ie 413 @ 64mhz
            U1BRGL = $9D
          $endif
        $else
          $if _device = _18F26K22 Or _device = _18F27K40
            Declare Hserial2_Baud  = 9600
            Declare Hserial2_TXSTA = 32                  ' 32 = bit5
          $else
            ' ***** work around for K42 - remove when issue resolved
            U2CON0 = %00110000                           ' BRGS=0 ABDEN=0 TXEN=1 RXEN=1 Mode=0000 (8bit async)
            U2CON1 = %10000000                           ' SerialEn=1 unimp=00 WUE=0 RXBIMD=0 unimp=0 BRKOVR=0 SENDB=0
            While U2FIFO.3 = 0                           ' RXIDL = bit 3 1 when idle
            Wend                                         '
            U2BRGH = 1                                   ' if idle set for 9600 ie 413 @ 64mhz
            U2BRGL = $9D
          $endif
        $endif
      $endif
    $endif

  $else                                                  ' bit bang, baby!
    $if D_BaudRate = 19200
      Dim D_SerBaud As 33                                ' 19200,n,1
    $elseif D_BaudRate = 38400
      Dim D_SerBaud As 7                                 ' 38400,n,1
    $else
      Dim D_SerBaud As  84                               ' 9600,n,1    anything else
    $endif

    Symbol PinOut = PGC                                  ' In this example ISCP data pins are being used for serial data
    Symbol PinIn = PGD                                   '
  $endif


'**************************** end of Device specific section ************************************


  GoTo D_JumpOver     ' <----------------------- from restart jump over this code

D_Dump:

  Wreg_Byte D_WregBack                                   ' ??? should I backup STATUS ???
  D_aBPF   = BPF
  D_aGEN   = GEN
  D_aGENH  = GENH
  D_aPP0   = PP0
  D_aPP0H  = PP0H
  D_aPP1   = PP1
  D_aPP1H  = PP1H
  D_aPP7   = PP7
  D_aPP7H  = PP7H
  D_Pbp#Var0 = Pbp#Var0

  ' jic
'  D_aPP2   = PP2
'  D_aPP2H  = PP2H
'  D_aPP4H  = PP4H
'  D_aPP6   = PP6
'  D_aPP6H  = PP6H
'  D_aPP8   = PP8
'  D_aPP8H  = PP8H

  $if D_SerialNum = 0
    D_aBPFH  = BPFH
    D_aGPR   = GPR
    D_aGEN3  = GEN3
    D_aGEN3H = GEN3H
    D_aPP3   = PP3
    D_aPP3H  = PP3H
    D_aPP4   = PP4
    D_aPP5   = PP5
    D_aPP5H  = PP5H
  $endif

D_SendSyncTO:
  D_AddressEnd = AddressOf(D_AddressBack)               ' D_AddressEnd is a constant
  If D_AddressStart = 0 Then
    D_AddressStart = AddressOf(BPF)
  End If
  D_AddressBack = D_AddressStart                        ' D_AddressStart may be set by caller, default to zero


' ********* S Y N C ********
D_SendSync:
  D_Data = D_Sync
  GoSub D_SendByte

  $if D_SerialNum = 1
    D_Data = HRSIn, {D_TimeOut, D_SendSync}           ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 2
    D_Data = HRSIn2, {D_TimeOut, D_SendSync}          ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, D_TimeOut, D_SendSync, [D_Data]    ' wait for ACK, NAK, Timeout or corrupt
  $endif

  If D_Data = D_SyncAck Then GoTo D_ExitSync            ' if waitack then exit

  $if _core = 16                                        ' Note - D_MT_Modify packets never tx'ed by host for 16F
    If D_Data = D_STX Then
      GoSub D_InBoundPacket                             ' if STX then try to get a packet
    Else
      ' what to do if char not SyncAck or stx?          ' ToDo - Wait 1 sec. flush the buffer
    End If
  $endif

  GoTo D_SendSync                                       ' if neither loop
D_ExitSync:


' ********* D A T A ********
  D_RetryCount = 0                                      ' setup retry
D_Data_Retry:
  If D_RetryCount > D_Retry Then GoTo D_SendSync
  Inc D_RetryCount
  D_Address = D_AddressBack                             ' set address (convienient for TO)
  D_DataLen = D_DataBlock                               ' is nec ?

D_SendData:
  D_AddressBack = D_Address                             ' save address

  $if D_CheckNulls = True
    D_TempWord = D_AddressEnd - D_Address
    If D_TempWord > D_DataBlock Then                    ' if remaining RAM less than data packet payload then don't check For contingous NULLs
      GoSub D_ScanForNull                               ' check for contingous NULLs
    End If
  $endif

  If D_Address = D_AddressBack Then                     ' test if NULL block sent (address has been adjusted to first non NULL)
    D_DataLen = D_DataBlock                             ' a NULL block packet was not sent to send a data packet
    GoSub D_SendDataPacket                              ' send a data packet
  End If

  $if D_SerialNum = 1
    D_Data = HRSIn, {D_TimeOut, D_Data_Retry}         ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 2
    D_Data = HRSIn2, {D_TimeOut, D_Data_Retry}        ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, D_TimeOut, D_Data_Retry, [D_Data]  ' Wait For ACK, NAK, Timeout or corrupt
  $endif

  If D_Data <> D_Ack Then D_Data_Retry                  ' fall thro if PC ack'ed
  D_RetryCount = 0                                      ' good tx so reset retry

  If D_Address < D_AddressEnd Then GoTo D_SendData      ' packet RX'ed check for overrun of specified block


' ********* E E P R O M  ********
  $if D_SendEEPROM = True

    D_AddressEnd = _eeprom                              ' NOTE D_AddressEnd is invalid from here
    $if D_LimitEEPROM = True
      If D_AddressEnd > 255 Then
        D_AddressEnd = 255                              ' testing only force a byte address
      End If
    $endif

    D_AddressBack = 0
    D_RetryCount = 0                                    ' setup retry
D_EEPROM_Retry:
    If D_RetryCount > D_Retry Then GoTo D_SendSync
    Inc D_RetryCount
    D_Address = D_AddressBack                           ' set address (convienient for TO)
    D_DataLen = D_EEPROMBlock                           ' is nec ?

D_EEPROMSend:
    D_AddressBack = D_Address

    D_TempWord = D_AddressEnd - D_Address
    If D_TempWord > D_EEPROMBlock Then                  ' if remaining EEPROM less than data packet payload then don't check For contingous NULLs
      GoSub D_ScanForFF                                 ' check for contiguous FF's
    End If

    If D_Address = D_AddressBack Then                   ' test if FF block sent (address has been adjusted to first non FF)
      D_DataLen = D_EEPROMBlock                         ' a FF block packet was not sent to send a data packet
      GoSub D_SendEEPROMPacket                          ' send a packet of EEPROM data
    End If

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_TimeOut, D_EEPROM_Retry}           ' wait for ACK, NAK, Timeout or corrupt
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_TimeOut, D_EEPROM_Retry}          ' wait for ACK, NAK, Timeout or corrupt
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_TimeOut, D_EEPROM_Retry, [D_Data]  ' Wait For ACK, NAK, Timeout or corrupt
    $endif

    If D_Data <> D_Ack Then D_EEPROM_Retry              ' fall thro if PC ack'ed
    D_RetryCount = 0                                    ' good tx so reset retry

    If D_Address < D_AddressEnd Then GoTo D_EEPROMSend  ' packet RX'ed check for overrun of specified block

    D_AddressEnd = AddressOf(D_Address)                 ' restore D_AddressEnd

  $endif


' *********** P O R T **********
  $if D_SendPort = True

    D_RetryCount = 0
D_PortRetry:
    If D_RetryCount > D_Retry Then GoTo D_SendSync
    Inc D_RetryCount
    GoSub D_Send_Port                                 ' send packet then wait for response

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_TimeOut, D_PortRetry}        ' wait for ACK, NAK, Timeout or corrupt
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_TimeOut, D_PortRetry}       ' wait for ACK, NAK, Timeout or corrupt
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_TimeOut, D_PortRetry, [D_Data]   ' Wait For ACK, NAK, Timeout or corrupt
    $endif

    If D_Data <> D_Ack Then D_PortRetry                 ' fall thro if PC ack'ed

  $endif


' *********** S F R **********
  $if D_SendSFR = True

    D_RetryCount = 0
D_SFRRetry:
    If D_RetryCount > D_Retry Then GoTo D_SendSync
    Inc D_RetryCount
    GoSub D_SendSFRPacket                              ' send packet then wait for response

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_TimeOut, D_SFRRetry}          ' wait for ACK, NAK, Timeout or corrupt
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_TimeOut, D_SFRRetry}         ' wait for ACK, NAK, Timeout or corrupt
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_TimeOut, D_SFRRetry, [D_Data]   ' Wait For ACK, NAK, Timeout or corrupt
    $endif

    If D_Data <> D_Ack Then D_SFRRetry                 ' fall thro if PC ack'ed

  $endif


' ********* E N D D A T A  ********
  D_RetryCount = 0
D_EndRetry:
  If D_RetryCount > D_Retry Then GoTo D_SendSync
  Inc D_RetryCount
  GoSub D_Send_EndData                                  ' send packet then wait for response

  $if D_SerialNum = 1
    D_Data = HRSIn, {D_TimeOut, D_EndRetry}           ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 2
    D_Data = HRSIn2, {D_TimeOut, D_EndRetry}          ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, D_TimeOut, D_EndRetry, [D_Data]    ' Wait For ACK, NAK, Timeout or corrupt
  $endif

  If D_Data <> D_Ack Then D_EndRetry                    ' fall thro if PC ack'ed


' ********* W A I T S Y N C ********
  If D_WaitFlag = 0 Then ExitDump                       ' no wait

  D_RetryCount = 0
D_SendWaitSyncTO:                                       ' timeout si inc retry
  Inc D_RetryCount                                      ' not untill host modified to send WaitNak
D_SendWaitSync:                                         ' loop here untill Ack'ed
  If D_RetryCount > D_Retry Then GoTo D_SendSyncTO

  D_Data = D_WaitSync
  GoSub D_SendByte

  $if D_SerialNum = 1
    D_Data = HRSIn, {D_TimeOut, D_SendWaitSyncTO}     ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 2
    D_Data = HRSIn2, {D_TimeOut, D_SendWaitSyncTO}    ' wait for ACK, NAK, Timeout or corrupt
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, D_TimeOut, D_SendWaitSyncTO, [D_Data] ' wait for ACK, NAK, Timeout or corrupt
  $endif

  If D_Data = D_WaitAck Then GoTo ExitDump              ' if waitack then exit

  $if _core = 16                                        ' Note - D_MT_Modify packets never tx'ed by host for 16F
    If D_Data = D_STX Then
      GoSub D_InBoundPacket                             ' if STX then try to get a packet
      D_RetryCount = 0
      GoTo D_SendWaitSync                               ' without this retry gets inc'ed
    End If
  $endif

  If D_Data = D_WaitNak Then                            ' if waitnak then wait and loop
    DelayMS D_TimeOut
    D_RetryCount = 0
  Else
    Inc D_RetryCount                                    ' if anything else inc retry
  End If

  'drop thro and purge for all inputs - WaitAck, WaitNak, other, InBoundPacket

  ' purge rx buffer
  $if D_SerialNum = 1
    D_Data = HRSIn, {0, D_SendWaitSync}
  $elseif D_SerialNum = 2
    D_Data = HRSIn2, {0, D_SendWaitSync}
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, 0, D_SendWaitSync, [D_Data] ' 
  $endif
  GoTo D_SendWaitSync                                   ' if neither loop

ExitDump:
  BPF      = D_aBPF
  GEN      = D_aGEN
  GENH     = D_aGENH
  PP0      = D_aPP0
  PP0H     = D_aPP0H
  PP1      = D_aPP1
  PP1H     = D_aPP1H
  PP7      = D_aPP7
  PP7H     = D_aPP7H
  Pbp#Var0 = D_Pbp#Var0

  ' jic
'  PP2   = D_aPP2
'  PP2H  = D_aPP2H
'  PP4H  = D_aPP4H
'  PP6   = D_aPP6
'  PP6H  = D_aPP6H
'  PP8   = D_aPP8
'  PP8H  = D_aPP8H

  $if D_SerialNum = 0
    BPFH     = D_aBPFH
    GPR      = D_aGPR
    GEN3     = D_aGEN3
    GEN3H    = D_aGEN3H
    PP3      = D_aPP3
    PP3H     = D_aPP3H
    PP4      = D_aPP4
    PP5      = D_aPP5
    PP5H     = D_aPP5H
  $endif
  Byte_Wreg D_WregBack

  Return

  Stop                                                  ' JIC

' *************************** S T A R T   O F   S U B S **************************************

'---------------------------------------------------------------------------------                                        RAM
' SUBROUTINE: D_ScanForNull                                                                                               RAM
' PURPOSE   : test for contiguous block of NULL starting at D_Address
' REQUIRES  : D_Address - in real world never 0
' RETURNS   : D_Datalen (if block found)
' DESTROYS  : D_Data, D_Counter and D_DataLen
'---------------------------------------------------------------------------------
D_ScanForNull:
  $if D_CheckNulls = True

    D_Counter = 0                                 ' initialise counter

    Do                                            ' head of loop, test condition at the end

      $if _core = 16
        D_Data = Ptr8(D_Address)                  ' get data
      $else
        If D_Address < $100 Then
          Bcf STATUS, IRP                         ' Set IRP to 0
        Else
          Bsf STATUS, IRP                         ' Set IRP to 1
        End If
        D_Data = D_Address                        '
        Movf D_Data, W                            ' Initialize FSR - workaround
        Movwf FSR
        Movf INDF, W                              ' read (D_Address) into W
        Wreg_Byte D_Data
      $endif

      If D_Data <> 0 Then Break
      If D_Address > D_AddressEnd Then Break      ' test for overrun
      If D_Counter = $FF Then Break
      Inc D_Counter                               ' adjust counter
      Inc D_Address                               ' adjust address

    Loop

    D_Address = D_AddressBack                     ' either way restore address
    If D_Counter > (D_DataBlock - 1) Then         ' throw back the little ones (both address and len correct)
      D_DataLen = D_Counter                       ' set len
      GoSub D_Send_NullBlock                      ' send block of contiguous NULLs
    End If

  $endif
  Return


'-----------------------------------------------------------------------------                                        RAM
' SUBROUTINE: D_Send_NullBlock                                                                                        RAM
' PURPOSE   : send contiguous nulls packet
' REQUIRES  : D_Address And D_DataLen
' RETURNS   : D_Address
' DESTROYS  : D_Data D_Counter
'-----------------------------------------------------------------------------
' STX, D_MT_NullBlock, AddressH, AddressL, length of NULL block, Pad0....Padn, CRC

D_Send_NullBlock:

  $if D_CheckNulls = True

    If D_DataLen = 0 Then Return
    D_Data = D_STX                                ' STX
    GoSub D_SendByte

    $if D_UseCRC = True
      GoSub D_CRC8_Start                          ' initialise CRC
    $endif

    D_Data = D_MT_NullBlock                       ' message type
    GoSub D_SendByteAndUpCRC

    D_Data = D_Address.HighByte                   ' Address hi
    GoSub D_SendByteAndUpCRC
    D_Data = D_Address.LowByte                    ' Address lo
    GoSub D_SendByteAndUpCRC

    D_Data = D_DataLen                            ' block size
    GoSub D_SendByteAndUpCRC

    For D_Counter = 1 To D_NullPad
      D_Data = D_Pad                              ' pad1 - padn
      GoSub D_SendByteAndUpCRC
    Next

    $if D_UseCRC = False
      D_Data = $55
    $else
      D_Data = D_CRC8_Comp                           ' CRC
    $endif
    GoSub D_SendByte

    D_Address = D_Address + D_DataLen             ' adjust address to end of NUL block

    ' purge rx buffer before last tx char leaves UART
    $if D_SerialNum = 1
      D_Data = HRSIn, {0, D_NullPurge}
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {0, D_NullPurge}
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, 0, D_NullPurge, [D_Data]   ' 
    $endif
D_NullPurge:

  $endif
  Return


'---------------------------------------------------------------------------------                                        RAM
' SUBROUTINE: D_SendDataPacket                                                                                            RAM
' PURPOSE   : Send data packet
' REQUIRES  : D_Address And D_DataLen
' RETURNS   : D_Address
' DESTROYS  : D_Data, D_DataLen, D_Counter
'---------------------------------------------------------------------------------
' STX, D_MT_Data, DataLen, AddressH, AddressL, Data0....Datan, CRC

D_SendDataPacket:

  D_TempWord = D_Address + D_DataLen
  If D_TempWord > D_AddressEnd Then  ' check for overun
    D_DataLen = D_AddressEnd - D_Address        ' if necesary adjust byte count
  End If
  If D_DataLen = 0 Then Return                  ' ensure that adjustment didn't bork it

  D_Data = D_STX                                ' STX
  GoSub D_SendByte

  $if D_UseCRC = True
    GoSub D_CRC8_Start                          ' initialise CRC
  $endif

  D_Data = D_MT_Data                            ' message type
  GoSub D_SendByteAndUpCRC

  D_Data = D_DataLen                            ' data byte count
  GoSub D_SendByteAndUpCRC

  D_Data = D_Address.HighByte
  GoSub D_SendByteAndUpCRC
  D_Data = D_Address.LowByte
  GoSub D_SendByteAndUpCRC
  '
  For D_Counter = 1 To D_DataBlock              ' loop getting data
    $if _core = 16
      D_Data = Ptr8(D_Address)                  ' trailing bytes may be overrun but DataCount will allow these to be ignored
    $else
      If D_Address < $100 Then
        Bcf STATUS, IRP                         ' Set IRP to 0
      Else
        Bsf STATUS, IRP                         ' Set IRP to 1
      End If

      D_Data = D_Address                        '
      Movf D_Data, W                            ' Initialize FSR - workaround
      Movwf FSR
      Movf INDF, W                              ' read (D_Address) into W
      Wreg_Byte D_Data
    $endif
    GoSub D_SendByteAndUpCRC
    Inc D_Address
  Next

  $if D_UseCRC = False
    D_Data = $55
  $else
    D_Data = D_CRC8_Comp                        ' CRC
  $endif
  GoSub D_SendByte

  ' purge rx buffer before last tx char leaves UART
  $if D_SerialNum = 1
    D_Data = HRSIn, {0, D_PacketPurge}
  $elseif D_SerialNum = 2
    D_Data = HRSIn2, {0, D_PacketPurge}
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, 0, D_PacketPurge, [D_Data] ' 
  $endif
D_PacketPurge:

  Return


'---------------------------------------------------------------------------------                                        EEPROM
' SUBROUTINE: D_ScanForFF                                                                                                 EEPROM
' PURPOSE   : test for contiguous block of FF starting at D_Address
' REQUIRES  : D_Address
' RETURNS   : D_Datalen (if block found)
' DESTROYS  : D_Data, D_Counter and D_DataLen
'---------------------------------------------------------------------------------
D_ScanForFF:

  $if D_SendEEPROM = True

    D_Counter = 0                                 ' initialise counter
    $if D_LimitEEPROM = True
      D_Data = ERead D_Address.LowByte
    $else
      D_Data = ERead D_Address
    $endif

    While D_Data = $FF                            ' test for $FF
      Inc D_Counter                               ' adjust counter
      Inc D_Address                               ' adjust address

      If D_Address > D_AddressEnd Then Break      ' test for overrun

      If D_Counter = $FF Then Break

      $if D_LimitEEPROM = True
        D_Data = ERead D_Address.LowByte
      $else
        D_Data = ERead D_Address
      $endif

    Wend

    D_Address = D_AddressBack                     ' either way restore address

    If D_Counter > (D_EEPROMBlock - 1) Then       ' throw back the little ones (both address and len correct)
      D_DataLen = D_Counter                       ' set len
      GoSub D_Send_FFBlock                        ' send block of contiguous $FF
    End If

  $endif
  Return


'-----------------------------------------------------------------------------                                        EEPROM
' SUBROUTINE: D_Send_FFBlock                                                                                          EEPROM
' PURPOSE   : send contiguous $FF packet
' REQUIRES  : D_Address And D_DataLen
' RETURNS   : D_Address
' DESTROYS  : D_Data D_Counter
'-----------------------------------------------------------------------------
' STX, D_MT_FFBlock, AddressH, AddressL, length of FF block, Pad0....Padn, CRC

D_Send_FFBlock:

  $if D_SendEEPROM = True

    If D_DataLen = 0 Then Return
    D_Data = D_STX                                ' STX
    GoSub D_SendByte

    $if D_UseCRC = True
      GoSub D_CRC8_Start                            ' initialise CRC
    $endif

    D_Data = D_MT_FFBlock                         ' message type
    GoSub D_SendByteAndUpCRC

    D_Data = D_Address.HighByte                   ' Address hi
    GoSub D_SendByteAndUpCRC
    D_Data = D_Address.LowByte                    ' Address lo
    GoSub D_SendByteAndUpCRC

    D_Data = D_DataLen                            ' block size
    GoSub D_SendByteAndUpCRC

    For D_Counter = 1 To D_FFPad
      D_Data = D_Pad                              ' pad1 - padn
      GoSub D_SendByteAndUpCRC
    Next

    $if D_UseCRC = False
      D_Data = $55
    $else
      D_Data = D_CRC8_Comp                           ' CRC
    $endif
    GoSub D_SendByte

    D_Address = D_Address + D_DataLen             ' adjust address to end of NUL block

    ' purge rx buffer before last tx char leaves UART
    $if D_SerialNum = 1
      D_Data = HRSIn, {0, D_FFPurge}
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {0, D_FFPurge}
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, 0, D_FFPurge, [D_Data]     ' 
    $endif
D_FFPurge:

  $endif
  Return


'---------------------------------------------------------------------------------                                        EEPROM
' SUBROUTINE: D_SendEEPROMPacket                                                                                          EEPROM
' PURPOSE   : Send EEPROM packet
' REQUIRES  : D_Address And D_DataLen
' RETURNS   : D_Address
' DESTROYS  : D_Data, D_DataLen, D_Counter
'---------------------------------------------------------------------------------
' STX, D_MT_EEPROM, DataLen, AddressH, AddressL, Data0....Datan, CRC

D_SendEEPROMPacket:

  $if D_SendEEPROM = True

    D_TempWord = D_Address + D_DataLen
    If D_TempWord > D_AddressEnd Then             ' check for overun
      D_DataLen = D_AddressEnd - D_Address        ' if necesary adjust byte count
    End If

    If D_DataLen = 0 Then Return                  ' ensure that adjustment didn't bork it

    D_Data = D_STX                                ' STX
    GoSub D_SendByte

    $if D_UseCRC = True
      GoSub D_CRC8_Start                          ' initialise CRC
    $endif

    D_Data = D_MT_EEPROM                          ' message type
    GoSub D_SendByteAndUpCRC

    D_Data = D_DataLen                            ' data byte count
    GoSub D_SendByteAndUpCRC

    D_Data = D_Address.HighByte                   ' send the start address
    GoSub D_SendByteAndUpCRC
    D_Data = D_Address.LowByte
    GoSub D_SendByteAndUpCRC

    For D_Counter = 1 To D_DataBlock                      ' loop getting data
      $if D_LimitEEPROM = True
        D_Data = ERead D_Address.LowByte
      $else
        D_Data = ERead D_Address
      $endif

      GoSub D_SendByteAndUpCRC
      Inc D_Address
    Next

    $if D_UseCRC = False
      D_Data = $55
    $else
      D_Data = D_CRC8_Comp                                ' CRC
    $endif
    GoSub D_SendByte

    ' purge rx buffer before last tx char leaves UART
    $if D_SerialNum = 1
      D_Data = HRSIn, {0, D_EEPROMPurge}
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {0, D_EEPROMPurge}
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, 0, D_EEPROMPurge, [D_Data]  ' 
    $endif
D_EEPROMPurge:

  $endif
  Return


'---------------------------------------------------------------------------------                                        PORT
' SUBROUTINE: D_Send_Port                                                                                                 PORT
' PURPOSE   : send port status packet
' REQUIRES  : nada
' RETURNS   : nada
' DESTROYS  : D_Data
' NOTES     :
'---------------------------------------------------------------------------------
' STX, D_MT_Port, TRISA, TRISB, TRISC, TRISD, TRISE, LATA, LATB, LATC, LATD ,LATE, Pad0...Padn, CRC

D_Send_Port:

  $if D_SendPort = True

    D_Data = D_STX                                  ' STX
    GoSub D_SendByte

    $if D_UseCRC = True
      GoSub D_CRC8_Start                            ' initialise CRC
    $endif

    D_Data = D_MT_Port                            ' message type
    GoSub D_SendByteAndUpCRC

    ' Note! assume all devices have TRISA, TRISB and TRISC

    D_Data = TRISA                                ' PortA config
    GoSub D_SendByteAndUpCRC
    D_Data = TRISB                                ' PortB config
    GoSub D_SendByteAndUpCRC
    D_Data = TRISC                                ' PortC config
    GoSub D_SendByteAndUpCRC

    $ifdef _TRISD
      D_Data = TRISD                              ' PortD config
    $else
      D_Data = 0                                  ' placeholder - if no TRISD on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    $ifdef _TRISE
      D_Data = TRISE                              ' PortE config
    $else
      D_Data = 0                                  ' placeholder - if no TRISE on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    $if _defined (_LATA)
      D_Data = LATA                               ' LATA value
    $elseif _defined (_PORTA)
      D_Data = PORTA                              ' use PORTA if no LATA
    $else
      D_Data = 0                                  ' placeholder - if no LATA or PORTA on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    $if _defined (_LATB)
      D_Data = LATB                               ' LATB value
    $elseif _defined (_PORTB)
      D_Data = PORTB                              ' use PORTB if no LATB
    $else
      D_Data = 0                                  ' placeholder - if no LATB or PORTB on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    $if _defined (_LATC)
      D_Data = LATC                               ' LATC value
    $elseif _defined (_PORTC)
      D_Data = PORTC                              ' use PORTC if no LATC
    $else
      D_Data = 0                                  ' placeholder - if no LATC or PORTC on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    $if _defined (_LATD)
      D_Data = LATD                               ' LATD value
    $elseif _defined (_PORTD)
      D_Data = PORTD                              ' use PORTD if no LATD
    $else
      D_Data = 0                                  ' placeholder - if no LATD or PORTD on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    $if _defined (_LATE)
      D_Data = LATE                               ' LATE value
    $elseif _defined (_PORTE)
      D_Data = PORTE                              ' use PORTE if no LATE
    $else
      D_Data = 0                                  ' placeholder - if no LATE or PORTE on test Device
    $endif
    GoSub D_SendByteAndUpCRC

    For D_Counter = 1 To D_PortPad
      D_Data = D_Pad                              ' pad1 - padn
      GoSub D_SendByteAndUpCRC
    Next

    $if D_UseCRC = False
      D_Data = $55
    $else
      D_Data = D_CRC8_Comp                        ' CRC
    $endif
    GoSub D_SendByte

    ' purge rx buffer before last tx char leaves UART
    $if D_SerialNum = 1
      D_Data = HRSIn, {0, D_PortPurge}
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {0, D_PortPurge}
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, 0, D_PortPurge, [D_Data]   ' 
    $endif
D_PortPurge:

  $endif
  Return


'---------------------------------------------------------------------------------                                        RAM
' SUBROUTINE: D_SendSFRPacket                                                                                            RAM
' PURPOSE   : Send SFR packet
' REQUIRES  : D_Address And D_DataLen
' RETURNS   : D_Address
' DESTROYS  : D_Data, D_DataLen, D_Counter
'---------------------------------------------------------------------------------
' STX, D_MT_SFR, AddressH, AddressL, Data0....Datan, CRC

D_SendSFRPacket:

  $if D_SendSFR = True
    D_TempWord = D_Address + D_DataLen
    If D_TempWord > D_AddressEnd Then  ' check for overun
      D_DataLen = D_AddressEnd - D_Address        ' if necesary adjust byte count
    End If
    If D_DataLen = 0 Then Return                  ' ensure that adjustment didn't bork it
  
    D_Data = D_STX                                ' STX
    GoSub D_SendByte
  
    $if D_UseCRC = True
      GoSub D_CRC8_Start                          ' initialise CRC
    $endif
  
    D_Data = D_MT_SFR                             ' message type
    GoSub D_SendByteAndUpCRC
  
    D_Data = D_Address.HighByte
    GoSub D_SendByteAndUpCRC
    D_Data = D_Address.LowByte
    GoSub D_SendByteAndUpCRC
    '
    For D_Counter = 1 To D_SFRBlock              ' loop getting data



      ' FUCK



      GoSub D_SendByteAndUpCRC
      Inc D_Address
    Next
  
    $if D_UseCRC = False
      D_Data = $55
    $else
      D_Data = D_CRC8_Comp                        ' CRC
    $endif
    GoSub D_SendByte
  
    ' purge rx buffer before last tx char leaves UART
    $if D_SerialNum = 1
       D_Data = HRSIn, {0, D_SFRPurge}
    $elseif D_SerialNum = 2
       D_Data = HRSIn2, {0, D_SFRPurge}
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, 0, D_SFRPurge, [D_Data] ' 
    $endif
D_SFRPurge:

  $endif
  Return


'---------------------------------------------------------------------------------                                        CONTROL
' SUBROUTINE: D_Send_EndData                                                                                              CONTROL
' PURPOSE   : send data complete and status packet
' REQUIRES  : nada
' RETURNS   : nada
' DESTROYS  : D_Data, D_Counter
'---------------------------------------------------------------------------------
' 16F STX, D_MT_EndData, D_DumpID, D_VerHi, D_VerLo, D_WaitFlag, D_AddStart, D_AddEnd, SourceID x14, D_Pad0..Pad6, CRC        25
' 18F STX, D_MT_EndData, D_DumpID, D_VerHi, D_VerLo, D_WaitFlag, D_AddStart, D_AddEnd, SourceID x14, DEVID x2, D_RomGoodFlag, D_Pad0..Pad3, CRC     28

D_Send_EndData:

  D_Data = D_STX                                ' STX                                         1
  GoSub D_SendByte

  $if D_UseCRC = True
    GoSub D_CRC8_Start                          ' initialise CRC
  $endif

  D_Data = D_MT_EndData                         ' message type                                2
  GoSub D_SendByteAndUpCRC

  D_Data = D_DumpID                             ' dump id                                     3
  GoSub D_SendByteAndUpCRC

  D_Data = D_VersionHigh                        ' version                                     4
  GoSub D_SendByteAndUpCRC

  D_Data = D_VersionLow                         ' version                                     5
  GoSub D_SendByteAndUpCRC

  D_Data = D_WaitFlag                           ' 1 Wait For host  0 Return At once           6
  GoSub D_SendByteAndUpCRC

  D_Data = D_AddressStart.HighByte              ' Start Address hi                            7
  GoSub D_SendByteAndUpCRC                      '
  D_Data = D_AddressStart.LowByte               ' Start Address lo                            8
  GoSub D_SendByteAndUpCRC                      '
                                                '
  D_Data = D_AddressEnd.HighByte                ' End Address hi                              9
  GoSub D_SendByteAndUpCRC                      '
  D_Data = D_AddressEnd.LowByte                 ' End Address lo                              10
  GoSub D_SendByteAndUpCRC                      '

  $if _core = 14
    D_AddressEnd = AddressOf(HexTag)            ' test for absence of CData string
    EEADR = D_AddressEnd.LowByte
    EEADRH = D_AddressEnd.HighByte
    EECON1 = %10000001                          ' read program memory
    Nop                                         ' two cycle stall
    Nop                                         ' EEDATH = 0x3F (erased) or 0x34 (Retlw)

    D_Counter = 0                               ' source ID                                    11 - 24
    While D_Counter < 14
      If EEDATH = 0x34 Then                          '
        D_Data = CRead (D_AddressEnd + D_Counter)    ' RETLW found, so do CRead
      Else
        D_Data = 0                                   ' no RETLW, so send placeholder
      End If
      GoSub D_SendByteAndUpCRC
      Inc D_Counter
    Wend

  $elseif _core = 16
    D_Counter = 0                                    ' source ID                               11 - 24
    While D_Counter < 14
      D_Data = CRead (AddressOf(HexTag) + D_Counter)
      GoSub D_SendByteAndUpCRC
      Inc D_Counter
    Wend
  $endif

'
' DEVID x2, D_RomGoodFlag, D_Pad0....Pad3, CRC  ' 32 - 28 = 4 bytes to pad      18F

  $if _core = 16
    ' 18F
    EECON1 = %11000000                          ' Access Configuration registers

    Asm
      Movlw Low (DEVID1)
      Movwf TBLPTRL
      Movlw High (DEVID1)
      Movwf TBLPTRH
      Movlw upper (DEVID1)
      Movwf TBLPTRU
    EndAsm

    Tblrd*+                                     ' perform table read with post-increment      25
    D_Data = TABLAT                             ' DEVID1
    GoSub D_SendByteAndUpCRC                    '
    Tblrd*                                      ' perform table read with post-increment      26
    D_Data = TABLAT                             ' DEVID2
    GoSub D_SendByteAndUpCRC                    '

    Asm                                         ' added 05/08/18 for silicon rev of new devices
      Movlw Low (DEVID1 - 2)
      Movwf TBLPTRL
    EndAsm

    Tblrd*+                                     ' perform table read with post-increment      27
    D_Data = TABLAT                             ' Revision ID    on 18F27K42
    GoSub D_SendByteAndUpCRC                    '
    Tblrd*+                                     ' perform table read with post-increment      28
    D_Data = TABLAT                             ' Revision ID    On 18F27K42
    GoSub D_SendByteAndUpCRC                    '

    TBLPTRU = 0                                 ' Proton expects this to be 0
    EECON1 = 0                                  ' Disable Access the Configuration registers

    D_Data = D_RomGoodFlag                      ' ROM pass fail compared to data from Host    29
    GoSub D_SendByteAndUpCRC

    For D_Counter = 1 To D_EndDataPad16         ' 05/08/18 reduced by 2 for Rev ID
      D_Data = D_Pad                            ' pad0 - pad3                                 30 - 31
      GoSub D_SendByteAndUpCRC
    Next

  $else       ' 16F

' D_Pad0....Pad6, CRC                           ' 32 - 25 = 7 bytes to pad   16F

    For D_Counter = 1 To D_EndDataPad14
      D_Data = D_Pad                            ' pad0 - pad6                                 25 - 31
      GoSub D_SendByteAndUpCRC
    Next
  $endif

  $if D_UseCRC = False
    D_Data = $55                                '                                             32
  $else
    D_Data = D_CRC8_Comp                        ' CRC
  $endif
  GoSub D_SendByte

  ' purge rx buffer before last tx char leaves UART
  $if D_SerialNum = 1
     D_Data = HRSIn, {0, D_EndPurge}
  $elseif D_SerialNum = 2
     D_Data = HRSIn2, {0, D_EndPurge}
  $elseif D_SerialNum = 0
    SerIn PinIn, D_SerBaud, 0, D_EndPurge, [D_Data] 
  $endif
D_EndPurge:

  Return


'----------------------------------------------------------------------                                        CONTROL
' SUBROUTINE: D_InBoundPacket                                                                                  CONTROL
' PURPOSE   : Recieve inbound packet, confirm validty and act upon
' REQUIRES  :
' RETURNS   :
' DESTROYS  : D_Data, D_DataLen, D_Address, D_Data1 - D_Data4
'----------------------------------------------------------------------
' STX, PacketLen, D_MT_Modify, DataLen, AddressH, AddressL, Data1, Data2, Data3, Data4, CRC         Variable modify
'
' DataLen=0 special case
' STX, PacketLen, D_MT_Modify, AddressH, AddressL, OrMask, AndMask, 0, 0, CRC                       Bit modify
'
' Read fom Address, OR with OrMask, AND with AndMask, Write to Address
'  PC will set either OrMask or AndMask  eg. AND %11110111 to clear, OR %00001000 to set
'
' STX, PacketLen, D_MT_Reset, 0, ResetFlagH, ResetFlagL, 0, 0, 0, 0, CRC                            Reset
'
' STX, PacketLen, D_MT_ROMEndData, DataLen, AddressH, AddressL, Byte1, Byte2, Byte3, Byte4, CRC     Source verfication
'  DataLen 2 or 4



' Note - 18F only - host will not tx D_MT_InBound packets for 16F

D_InBoundPacket:

  $if _core = 16

    D_RxFailFlag = False
    D_Data = D_STX                                        ' add STX into crc
    $if D_UseCRC = True
      GoSub D_CRC8_Start                                  ' initialise CRC
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get packet length (not currently used)
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get message type
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_MessageType = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get data length
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_DataLen = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get address - high byte
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_Address.HighByte = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get address - low byte
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_Address.LowByte = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get data 1
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_Data1 = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get data 2
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_Data2 = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get data 3
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif

    D_Data3 = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get data 4
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    D_Data4 = D_Data
    $if D_UseCRC = True
      GoSub D_CRC8_Add
    $endif

    $if D_SerialNum = 1
      D_Data = HRSIn, {D_RxTimeOut, D_RxPacketTo}         ' get crc
    $elseif D_SerialNum = 2
      D_Data = HRSIn2, {D_RxTimeOut, D_RxPacketTo}        ' 
    $elseif D_SerialNum = 0
      SerIn PinIn, D_SerBaud, D_RxTimeOut, D_RxPacketTo, [D_Data]  ' 
    $endif
    $if D_UseCRC = True
      If D_Data <> D_CRC8_Comp Then
        D_RxFailFlag = True
      End If
    $endif


    Select D_MessageType

    Case D_MT_ROMEndData

      Select D_DataLen
      Case 2
        If D_Data2 <> cPtr8(D_Address) Then D_RxFailFlag = True         ' note the byte swap in both cases
        Inc D_Address
        If D_Data1 <> cPtr8(D_Address) Then D_RxFailFlag = True

      Case 4
        If D_Data2 <> cPtr8(D_Address) Then D_RxFailFlag = True
        Inc D_Address
        If D_Data1 <> cPtr8(D_Address) Then D_RxFailFlag = True
        Inc D_Address
        If D_Data4 <> cPtr8(D_Address) Then D_RxFailFlag = True
        Inc D_Address
        If D_Data3 <> cPtr8(D_Address) Then D_RxFailFlag = True

      Case Else
        D_RxFailFlag = True

      End Select

      If D_RxFailFlag = False Then
        D_RomGoodFlag = True
        D_Data = D_Ack                                      ' D_Ack
      Else
        D_RomGoodFlag = False
        D_Data = D_Nak                                      ' D_Nak
      End If
      GoSub D_SendByte
      Return

    Case D_MT_Reset
      If D_DataLen <> 0 Then D_RxFailFlag = True
      If D_Data1 <> 0 Then D_RxFailFlag = True
      If D_Data2 <> 0 Then D_RxFailFlag = True
      If D_Data3 <> 0 Then D_RxFailFlag = True
      If D_Data4 <> 0 Then D_RxFailFlag = True
      If D_Address.HighByte <> D_ResetFlagH Then D_RxFailFlag = True
      If D_Address.LowByte <> D_ResetFlagL Then D_RxFailFlag = True

      If D_RxFailFlag = False Then
        D_Data = D_Ack
        GoSub D_SendByte
        DelayMS 10
        Reset
      Else
        D_Data = D_Nak
        GoSub D_SendByte
      End If
      Return

    Case D_MT_Modify
      If D_Address < D_AddressStart Then
        D_RxFailFlag = True
      End If
        'If D_Address > D_AddressEnd Then                   ' need to add data address check !!!! allow Start-End and HeapStart-HeapEnd
        '  D_RxFailFlag = True
        'End If
      If D_DataLen > 5 Then
        D_RxFailFlag = True
      End If

      If D_RxFailFlag = False Then
        If D_DataLen = 0 Then                               '
          D_Data = Ptr8(D_Address)                          ' get data
          D_Data = D_Data | D_Data1                         ' OR
          D_Data = D_Data & D_Data2                         ' AND
          Ptr8(D_Address) = D_Data
        Else
          Ptr8(D_Address) = D_Data1                         ' always data1
          If D_DataLen > 1 Then                             ' 2?
            Inc D_Address
            Ptr8(D_Address) = D_Data2                       ' do 2
          End If
          If D_DataLen > 2 Then                             ' 3?
            Inc D_Address
            Ptr8(D_Address) = D_Data3                       ' do 3
            Inc D_Address                                   ' must be 4
            Ptr8(D_Address) = D_Data4                       ' do 4
            Inc D_Address
          End If
        End If

        D_Data = D_Ack                                      ' D_Ack
      Else
        D_Data = D_Nak                                      ' D_Nak
      End If
      GoSub D_SendByte
      Return

    Case Else
      D_Data = D_NotSupported
      GoSub D_SendByte
      Return

    End Select

    Return                                                  ' jic

D_RxPacketTo:

    D_Data = D_Nak                                          ' what to do if we timeout?
    GoSub D_SendByte

  $endif
  Return                                                    ' JIC it gets called accidentally


'----------------------------------------------------------------------                                        UTILITY
' SUBROUTINE: D_SendByte                                                                                       UTILITY
' PURPOSE   : Send a byte
' REQUIRES  : D_Data
' RETURNS   :
' DESTROYS  :
'----------------------------------------------------------------------
D_SendByte:
  ' compile time switch between UART and bit bang
  $if D_SerialNum = 1
    HRSOut D_Data
  $elseif D_SerialNum = 2
    HRSOut2 D_Data
  $elseif D_SerialNum = 0
    SerOut PinOut, D_SerBaud, [D_Data]              ' bit bang, baby
  $endif
  Return


'----------------------------------------------------------------------                                        UTILITY
' SUBROUTINE: D_SendByteAndUpCRC                                                                               UTILITY
' PURPOSE   : Send a byte and update the running CRC
' REQUIRES  : D_Data
' RETURNS   :
' DESTROYS  : D_Data
'----------------------------------------------------------------------
D_SendByteAndUpCRC:
  GoSub D_SendByte
  $if D_UseCRC = True
    GoSub D_CRC8_Add                                  ' add to running CRC
  $endif
  Return


'----------------------------------------------------------------------                                        UTILITY
' SUBROUTINE: D_CRC8  (note! dual entry)                                                                       UTILITY
' PURPOSE   : computes a running CRC
' REQUIRES  : D_CRC8_Comp
' RETURNS   :
' NOTES     : originally coded by Colin O'Flynn - Copyright (c) 2002 in C
'----------------------------------------------------------------------
  $if D_UseCRC = True
D_CRC8_Start:
    D_CRC8_Comp = 0
D_CRC8_Add:
    D_CRC8_bit_counter = 8
    While D_CRC8_bit_counter > 0
      D_CRC8_feedback_bit = (D_CRC8_Comp ^ D_Data) & $01
      If D_CRC8_feedback_bit = $01 Then
        D_CRC8_Comp = D_CRC8_Comp ^ D_CRC8POLY
      EndIf
      D_CRC8_Comp = (D_CRC8_Comp >> 1) & $7F
      If D_CRC8_feedback_bit = $01 Then
        D_CRC8_Comp = D_CRC8_Comp | $80
      EndIf
      D_Data = D_Data >> 1
      D_CRC8_bit_counter = D_CRC8_bit_counter - 1
    Wend
    Return
  $endif

' *************************** S T A R T  O F   M A C R O S **************************************

  $define Dump(pValue) '
  D_DumpID = pValue    '
  D_WaitFlag = 0       '
  GoSub D_Dump

  $define DumpW(pValue) '
  D_DumpID = pValue     '
  D_WaitFlag = 1        '
  GoSub D_Dump

' *************************** E N D   O F   M A C R O S **************************************

D_JumpOver:
  ' compile time messages to display option selection
  $if EnableSnapShot = True
    $if D_SerialNum <> 0
      $if _uart = 2
        $if D_SerialNum = 1
          $SendWarning "This device has 2 UARTS, are you sure you want to use the first one for SnapShot data?"
        $elseif D_BaudRate = 115200
          $SendHint "SnapShot data will be generated by UART2 at 115200 baud"
        $elseif D_BaudRate = 57600
          $SendHint "SnapShot data will be generated by UART2 at 57600 baud"
        $elseif D_BaudRate = 9600
          $SendHint "SnapShot data will be generated by UART2 at 9600 baud"
        $endif
      $else                                                                   ' _uart = 1
        $if D_BaudRate = 115200
          $SendHint "SnapShot data will be generated by UART1 at 115200 baud"
        $elseif D_BaudRate = 57600
          $SendHint "SnapShot data will be generated by UART1 at 57600 baud"
        $elseif D_BaudRate = 9600
          $SendHint "SnapShot data will be generated by UART1 at 9600 baud"
        $endif
      $endif
    $else
      $SendHint "Bit banging will be used to generated SnapShot data at 9600 baud"
    $endif

    $define D_AllDefault False

    $if D_UseCRC = True
      $if D_CheckNulls = True
        $if D_SendEEPROM = True
          $if D_SendPort = True
            $define D_AllDefault True
            $SendHint "SnapShot is configured for default CRC, Checknulls, EEPROM and Port behavior"
          $endif
        $endif
      $endif
    $endif

    $if D_AllDefault = False
      $SendHint "Verify SnapShot settings for CRC, Checknulls, EEPROM and Port data, at least one has non-standard value"
    $endif
  $endif

'end of SnapShot.inc
